/*
 * $Header: /home/puhuri/Tutkimus/src/timer/RCS/sender.c,v 1.3 1997/02/19 12:46:22 puhuri Exp puhuri $
 *
 * $Log: sender.c,v $
 * Revision 1.3  1997/02/19 12:46:22  puhuri
 * Ekojen mittausten versio
 *
 * Revision 1.2  1996/11/22 08:29:35  puhuri
 * Muutettu uuteen rakenteesetn.
 *
 * Revision 1.1  1996/11/05 19:00:28  puhuri
 * Initial revision
 *
 *
 *
 *
 */
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <arpa/inet.h>
#include <signal.h>
#include <setjmp.h>
#include "sched_tst.h"


#ifdef RCS_INFO
static char rcs_id[] = "$Id: sender.c,v 1.3 1997/02/19 12:46:22 puhuri Exp puhuri $";
#endif

static int sender_child (struct conf_info *conf, int mynro);
#ifdef SECURE_SIGPIPE
static void pipe_closed (int sig);
static jmp_buf env;
#endif

int sender (struct conf_info *conf)
{
  static int i = 0;
  pid_t pid;
  int try = 40;

  if (i >= conf->wprogs)
    return (0);			/* all started */

  conf->childs_running++;
  do {
    if ((pid = fork()) == 0) {	/* lapsiprosessi */
      sender_child (conf, i);	/* tehdn mit halutaan */
      exit (OK);			/* palataan */
    }
  } while (pid < 0 && (errno == EAGAIN || errno == EINTR) && -- try > 0);

  if (pid < 0) {		/* fork eponnistui */
    conf->childs_running--;
    PERROR ("fork: ", ERR_FORK_FAILED);
  }
  add_child_pid (conf, pid);
  i++;
  return (1);
}

static int sender_child (struct conf_info *conf, int mynro)
{
  struct log_entry *log;
  char  *read_buf,  *write_buf;
  int inb, outb, j, i = 0;
  int fd, rfd;			/* listen fd, read fd */
  struct sockaddr_in my_addr;
  struct sockaddr_in other_addr;
  int soa_len;
  struct hostent *he;
  int write_times = conf->write_times;
  int write_size = conf->write_size;
  int before_writes = 0;
  int before_bytes = 0;

  conf->wait_for_signal = 1;
  signal (SIGUSR1, child_wait_sig);

  if ((read_buf = malloc(write_size)) == NULL) {
    PERROR ("malloc: ", ERR_NO_MEMORY);
  }
  if ((write_buf = malloc(write_size)) == NULL) {
    PERROR ("malloc: ", ERR_NO_MEMORY);
  }

  if ((log = get_dirty_mem (ENTRY_PER_LOOP * write_times * 
		     sizeof(struct log_entry))) == NULL) {
    PERROR ("malloc: ", ERR_NO_MEMORY);
  }

  if ((rfd = open (conf->in_file, O_RDONLY)) < 0) {
    fprintf (stderr, "Filename: %s\n", conf->in_file);
    PERROR ("open: ", ERR_FILEOPEN);
  }
  do {
    off_t filepos;

    if ((filepos = lseek (rfd, 0, SEEK_END)) < 0 ||
	lseek (rfd, mynro * (filepos / (1 + conf->wprogs)), SEEK_SET) < 0)
      PERROR ("lseek: ", ERR_LSEEK);
  } while (0);

  if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    PERROR("socket:", ERR_SOCKET);
  }

  if ((he = gethostbyname ( conf->host )) == NULL) {
    PERROR ("gethostbyname: ", ERR_GETHOSTBYNAME);
  }

  other_addr.sin_port = htons (conf->port);
  other_addr.sin_family = AF_INET;
  other_addr.sin_addr = *((struct in_addr *)he->h_addr);

  if (connect (fd, (struct sockaddr *)&other_addr, 
	       sizeof(struct sockaddr_in)) < 0) {
    PERROR ("connect: ", ERR_CONNECT);
  }

  soa_len = sizeof(struct sockaddr_in);
  if (getsockname (fd, (struct sockaddr *)&my_addr, &soa_len)) {
    PERROR ("getsockname", ERR_GETSOCKNAME);
  }

  if (kill (getppid(), SIGUSR1)) {
    PERROR ("kill", ERR_KILL);
  }

  DPRINTF ("I (%d) signaled parent %d", getpid(), getppid());
  while (conf->wait_for_signal)
    sleep (1);

  DPRINTF ("%d: write_times=%d write_size=%d", getpid(),
	   write_times, write_size);

#ifdef SECURE_SIGPIPE
  if (setjmp(env) == 0) {
    signal (SIGPIPE, pipe_closed); /* ei voida asettaa aiemmin */
#endif
    i = 0;
    while (i < write_times) {
      while ((inb = read (rfd, read_buf, write_size)) > 0) {
	STORE_LOG (0, inb);
	for (j = 0; j < inb; j++)
	  write_buf[j] = read_buf[j];
	STORE_LOG (1, inb);
	if ((outb = write (fd, write_buf, inb)) < 0) {
	  PERROR ("write", ERR_WRITE);
	}
	STORE_LOG (2, outb);
	if (outb == 0)		/* kirjoitus suljettu */
	  break;
	if (timercmp(&log[ENTRY_PER_LOOP * i + 2].timeval, 
		     &(conf->go_time), <)) {
	  before_bytes += outb;	/* paljonkos lhetetn ennen */
	  before_writes++;	
	  continue;
	}

	if (timercmp(&log[ENTRY_PER_LOOP * i + 2].timeval, 
		     &(conf->stop_time), >))
	  break;
	if(++i >= write_times)
	  break;			/* valvotaan, ettei ylity */
      }
      if (inb == 0){		/* EOF */
	if (lseek (rfd, 0, SEEK_SET) < 0) {
	  PERROR ("lseek: ", ERR_LSEEK);
	}
      }
      if (timercmp(&log[ENTRY_PER_LOOP * i + 2].timeval, 
		   &(conf->stop_time), >))
	break;
    }
    close (fd);
    close (rfd);
#ifdef SECURE_SIGPIPE
  } else {
    fprintf (stderr, "%d: i == %d\n", getpid(), i);
  }
#endif

  if (conf->usleep_end > 0){		/* maybe sleep some... */
#ifdef _HPUX_SOURCE		/* HPUX: ei usleep */
    sleep (conf->usleep_end / i10E6 ? : 1);
#else
    usleep (conf->usleep_end);
#endif
  }

  printout_time_log (conf, log, (i - 1) * ENTRY_PER_LOOP, mynro, 
		     ID_CHAR_SEND);
  fprintf (stderr, 
	   "%d:%c mynro = %d port = %d: before writes = %d (%d bytes)\n",
	   getpid(), ID_CHAR_SEND, mynro, ntohs (my_addr.sin_port),
	   before_writes, before_bytes);
  exit (OK);
}

#ifdef SECURE_SIGPIPE
static void pipe_closed (int sig) {
  fprintf (stderr, "%d: Got signal %d\n", getpid(), sig);
  longjmp(env, sig);
}
#endif
